from numpy import *
from numpy import linalg, corrcoef
from scipy import special
from sklearn.neighbors import KDTree

def AIS (target, kHistory, kTau, k):
    
    """
	     Computes active information storage of a variable
         
        Args: 
            target: ndarray
                array of samples
            kHistory: int
                embedding vector length
            kTau: int
            	 embedding vector delay length
            k: int
                number of nearest neighbours in k-NN search
        
        Returns: float64
                active information storage value
         
	 """
    
    X=matrix(target[(1+kHistory*kTau-kTau):len(target)]).T
    Z=zeros((len(target)-(1+kHistory*kTau-kTau),kHistory))
    
     
    for i in range((1+kHistory*kTau-kTau),len(target)):
        Z[i-(1+kHistory*kTau-kTau),:]=target[i-(1+kHistory*kTau-kTau):i][((1+kHistory*kTau-kTau)%kTau+(kTau-1))%kTau::kTau]
    
    Z=matrix(Z)
    XZ=hstack((X,Z))
    
    #k-NN search
    treexz = KDTree(XZ, metric="chebyshev")
    treex = KDTree(X, metric="chebyshev")
    treez = KDTree(Z, metric="chebyshev")
    
    epsilon=nextafter(treexz.query(XZ, k=k+1)[0][:,k],0)
    sumx=treex.query_radius(X, r=epsilon, count_only=True)
    sumz=treez.query_radius(Z, r=epsilon, count_only=True)
                   
    sumDigammax=special.digamma(sumx)
    sumDigammaz=special.digamma(sumz)
    
    return mean(-sumDigammaz-sumDigammax) + special.digamma(k) + special.digamma(len(sumDigammax))

def TE (source, target, kHistory, kTau, lHistory, lTau, u, k):
    
    """
        Calculates Transfer entropy matrix of the whole system.	
	
    	Args:
		  source: ndarray
    			array of samples of the source variable
        target: ndarray
    			array of samples of the target variable
		  kHistory: int 
    			target history length
		  kTau: int
    			target delay length.
		  lHistory: int 
    			source history length
		  lTau: int
    			source delay length
		  u: int
    			source-target lag
		  k: int
    			number of nearest neighbours in k-NN search

    	Returns: float64
			transfer entropy value
	"""
    
    X=matrix(target[max(u+lHistory*lTau-lTau,1+kHistory*kTau-kTau):len(target)]).T
    Y=zeros((len(target)-max(u+lHistory*lTau-lTau,1+kHistory*kTau-kTau),lHistory))
    Z=zeros((len(target)-max(u+lHistory*lTau-lTau,1+kHistory*kTau-kTau),kHistory))
    
    for i in range(max(u+lHistory*lTau-lTau,1+kHistory*kTau-kTau),len(target)):
        Y[i-max(u+lHistory*lTau-lTau,1+kHistory*kTau-kTau),:]=source[i-(u+lHistory*lTau-lTau):i-(u-1)][((u+lHistory*lTau-lTau)%lTau+(lTau-u))%lTau::lTau]
        Z[i-max(u+lHistory*lTau-lTau,1+kHistory*kTau-kTau),:]=target[i-(1+kHistory*kTau-kTau):i][((1+kHistory*kTau-kTau)%kTau+(kTau-1))%kTau::kTau]
    
    XYZ=hstack((X,Y,Z))
    XZ=XYZ[:,r_[0,lHistory+1:lHistory+1+kHistory]]
    YZ=XYZ[:,r_[1:lHistory+1,lHistory+1:lHistory+1+kHistory]]
    Z=XYZ[:,r_[lHistory+1:lHistory+1+kHistory]]
    
    #k-NN search
    treexyz = KDTree(XYZ, metric="chebyshev")
    treexz = KDTree(XZ, metric="chebyshev")
    treeyz = KDTree(YZ, metric="chebyshev")
    treez = KDTree(Z, metric="chebyshev")
    
    epsilon=nextafter(treexyz.query(XYZ, k=k+1)[0][:,k],0)
    sumxz=treexz.query_radius(XZ, r=epsilon, count_only=True)
    sumyz=treeyz.query_radius(YZ, r=epsilon, count_only=True)
    sumz=treez.query_radius(Z, r=epsilon, count_only=True)
               
    sumDigammaxz=special.digamma(sumxz)
    sumDigammayz=special.digamma(sumyz)
    sumDigammaz=special.digamma(sumz)
     
    return mean(sumDigammaz-sumDigammayz-sumDigammaxz) + special.digamma(k)

def MC(Win, W,data,memory):
    
   """
    computes memory capacity of the network with single input and single output	
	
    	Args:
		  Win: ndarray
    			input weight matrix
		  W: ndarray 
    			reservoir weight matrix
		  data: ndarray
    			driving input signal
        memory: int
            memory lenght to be computed

    	Returns: float64
			transfer entropy value
	"""
    
    resize=Win.shape[0]
    insize=1
    outsize=1
    initLen=100
    trainLen=1000
    testLen=2000
    a=-1
    b=1
    
    u=data
    Wout=zeros((outsize,resize))
    x=zeros(resize)
    Xtrain=zeros((resize,trainLen))
    Xtest=zeros((resize,testLen))
    Ytest=zeros((memory,testLen))
    Ytrain=zeros((memory,trainLen))
        
    for t in range(memory+trainLen+1):    
        x=tanh(dot(W,x) +dot(Win,squeeze(u[t])))
            
        if(t>=memory+1):            
            Xtrain[:,t-memory-1]=x
        
    for t in  range(memory):   
        Ytrain[t,:]= squeeze(u[memory-t:memory+trainLen-t])
                   
    Wout=dot(Ytrain,linalg.pinv(Xtrain))
    Ytest=zeros((memory,testLen))           
       
    for t in range (testLen):   
        x=tanh(dot(W,x) +dot(Win,squeeze(u[memory+trainLen+1+t])))
        Xtest[:,t-memory-1]=x
        Ytest[:,t]=dot(Wout,x)
        
    MC=zeros(memory)
    for i in range(memory):
        k=i
        MC[i]=power(corrcoef(Ytest[i,:],squeeze(u[memory+trainLen-k:memory+trainLen+testLen-k]))[0,1],2)         
        
    return sum(MC), Xtest

def LE(W, WI, data, gamma0=10**-12, iterations=2000):
    
   """
    computes maximum Lyapunov exponent of the ESN	
	
    	Args:
  		  W: ndarray 
    			reservoir weight matrix
		  Win: ndarray
    			input weight matrix
		  data: ndarray
    			driving input signal
        gamma0: float64
            size of initial perturbation

    	Returns: float64
			maximum Lyapunov exponent value
	"""
    
    q=size(WI)
    lambdasum = 0
    for noda in range(q):
        X = zeros(q)
        X2 = zeros(q)
        X2[noda] += gamma0
        lambdas = zeros(1000)
        for it in range(iterations):
            X = tanh(dot(W, X) + dot(WI, data[it]))
            X2 = tanh(dot(W, X2) + dot(WI, data[it]))
            if (it>=1000):
                difr = X2 - X
                gammaK = sqrt(difr.dot(difr))
                X2 = X + difr * (gamma0 / gammaK)
                lambdas[it-1000] = log(gammaK / gamma0)
                    
        lambdasum += average(lambdas)
        
    return lambdasum / q
    
def ragwitz(target,kHistoryMax,kTauMax,k):
    
   """
    Ragwitz-Kantz embedding criterion grid search	
	
    	Args:
		  target: ndarray
    			array of samples
		  kHistoryMax: int 
    			maximum value of embedding vector lenght to be tested
		  kTauMax: int
    			maximum value of embedding vector delay to be tested
		  k: int
    			number of nearest neighbours to be used in epsylon estimation in locally constant predictor

    	Returns: 2-tuple
			 pair (k,tau) that minimized the locally constant prediction error found by grid search 
             k - embedding vector lenght
             tau - ebedding vector delay
	"""
    
    def locally_const_predictor(target, kHistory, kTau, k):
        #search space preparation
        Z=zeros((len(target)-(1+kHistory*kTau-kTau),kHistory))
        X=target[(1+kHistory*kTau-kTau):len(target)]
           
        for i in range((1+kHistory*kTau-kTau),len(target)):
            Z[i-(1+kHistory*kTau-kTau),:]=target[i-(1+kHistory*kTau-kTau):i][((1+kHistory*kTau-kTau)%kTau+(kTau-1))%kTau::kTau]
        
        Z=matrix(Z)
        
        #k-NN search
        treez = KDTree(Z, metric="chebyshev")
            
        kNNz=treez.query(Z, k=k+1)[1]
        prediction=sum(X[kNNz[:,1:(k+1)]],axis=1)/k
                      error=mean((prediction-X)**2)
    
        return error

    kHistory=1
    kTau=1
    kHistoryBest=1
    kTauBest=1
    bestPredictionError=locally_const_predictor(target,kHistory,kTau,k)
    
    for i in range(kHistoryMax-1):
        kHistory+=1
        for j in range(kTauMax):
            kTau=j+1
            predictionError=locally_const_predictor(target,kHistory,kTau,k)
            if(predictionError<bestPredictionError):
                bestPredictionError=predictionError
                kHistoryBest=kHistory
                kTauBest=kTau

    return  kHistoryBest, kTauBest  

def entropy(data,k):
    
    """
	  computes differential entropy using Kozachenko-Leonenko entropy estimation
         
        Args: 
            data: ndarray
                array of samples
            k: int
                number of nearest neighbours in k-NN search
        
        Returns: float64
                differential entropy value
         
	 """
    
    X=matrix(data).T
    treex = KDTree(X, metric="chebyshev")       
    kNNx=treex.query(X, k=k+1)[0][:,k]    
       
    return mean(log(2*kNNx)) - special.digamma(k) + special.digamma(len(kNNx))

def main():
    print("ESNtoolbox")

if __name__ == '__main__':
    main()